@@ -33,10 +33,11 @@ from typing import TYPE_CHECKING
 
 from libqtile import configurable, hook, utils
 from libqtile.bar import Bar
-from libqtile.command.base import CommandObject
+from libqtile.command.base import CommandObject, expose_command
 from libqtile.log_utils import logger
 
 if TYPE_CHECKING:
+    import re
     from typing import Any, Callable, Iterable
 
     from libqtile.backend import base
@@ -44,6 +45,7 @@ if TYPE_CHECKING:
     from libqtile.command.base import ItemT
     from libqtile.core.manager import Qtile
     from libqtile.group import _Group
+    from libqtile.layout.base import Layout
     from libqtile.lazy import LazyCall
 
 
@@ -60,19 +62,28 @@ class Key:
     key:
         A key specification, e.g. ``"a"``, ``"Tab"``, ``"Return"``, ``"space"``.
     commands:
-        A list :class:`LazyCall` objects to evaluate in sequence upon keypress.
+        One or more :class:`LazyCall` objects to evaluate in sequence upon keypress. Multiple
+        commands should be separated by commas.
     desc:
         Description to be added to the key binding. (Optional)
-
+    swallow:
+        Configures when we swallow the key binding. (Optional)
+        Setting it to False will forward the key binding to the focused window after the commands have been executed.
     """
 
     def __init__(
-        self, modifiers: list[str], key: str, *commands: LazyCall, desc: str = ""
+        self,
+        modifiers: list[str],
+        key: str,
+        *commands: LazyCall,
+        desc: str = "",
+        swallow: bool = True,
     ) -> None:
         self.modifiers = modifiers
         self.key = key
         self.commands = commands
         self.desc = desc
+        self.swallow = swallow
 
     def __repr__(self) -> str:
         return "<Key (%s, %s)>" % (self.modifiers, self.key)
@@ -99,6 +110,13 @@ class KeyChord:
     name:
         A string to name the chord. The name will be displayed in the Chord
         widget.
+    desc:
+        A string to describe the chord. This attribute is not directly used by Qtile
+        but users may want to access this when creating scripts to show configured
+        keybindings.
+    swallow:
+        Configures when we swallow the key binding of the chord. (Optional)
+        Setting it to False will forward the key binding to the focused window after the commands have been executed.
     """
 
     def __init__(
@@ -108,6 +126,8 @@ class KeyChord:
         submappings: list[Key | KeyChord],
         mode: bool | str = False,
         name: str = "",
+        desc: str = "",
+        swallow: bool = True,
     ):
         self.modifiers = modifiers
         self.key = key
@@ -116,6 +136,7 @@ class KeyChord:
         self.submappings = submappings
         self.mode = mode
         self.name = name
+        self.desc = desc
 
         if isinstance(mode, str):
             logger.warning(
@@ -126,6 +147,7 @@ class KeyChord:
             )
             self.name = mode
             self.mode = True
+        self.swallow = swallow
 
     def __repr__(self) -> str:
         return "<KeyChord (%s, %s)>" % (self.modifiers, self.key)
@@ -159,6 +181,9 @@ class Drag(Mouse):
         A list :class:`LazyCall` objects to evaluate in sequence upon drag.
     start:
         A :class:`LazyCall` object to be evaluated when dragging begins. (Optional)
+    warp_pointer:
+        A :class:`bool` indicating if the pointer should be warped to the bottom right of the window
+        at the start of dragging. (Default: `False`)
 
     """
 
@@ -168,9 +193,11 @@ class Drag(Mouse):
         button: str,
         *commands: LazyCall,
         start: LazyCall | None = None,
+        warp_pointer: bool = False,
     ) -> None:
         super().__init__(modifiers, button, *commands)
         self.start = start
+        self.warp_pointer = warp_pointer
 
     def __repr__(self) -> str:
         return "<Drag (%s, %s)>" % (self.modifiers, self.button)
@@ -269,6 +296,42 @@ class EzKey(EzConfig, Key):
         super().__init__(modkeys, key, *commands, desc=desc)
 
 
+class EzKeyChord(EzConfig, KeyChord):
+    """
+    Define a key chord using the Emacs-like format.
+
+    Parameters
+    ==========
+    keydef:
+        The Emacs-like key specification, e.g. ``"M-S-a"``.
+    submappings:
+        A list of :class:`Key` or :class:`KeyChord` declarations to bind in this chord.
+    mode:
+        Boolean. Setting to ``True`` will result in the chord persisting until
+        Escape is pressed. Setting to ``False`` (default) will exit the chord once
+        the sequence has ended.
+    name:
+        A string to name the chord. The name will be displayed in the Chord
+        widget.
+    desc:
+        A string to describe the chord. This attribute is not directly used by Qtile
+        but users may want to access this when creating scripts to show configured
+        keybindings.
+
+    """
+
+    def __init__(
+        self,
+        keydef: str,
+        submappings: list[Key | KeyChord],
+        mode: bool | str = False,
+        name: str = "",
+        desc: str = "",
+    ):
+        modkeys, key = self.parse(keydef)
+        super().__init__(modkeys, key, submappings, mode, name, desc)
+
+
 class EzClick(EzConfig, Click):
     """
     Bind commands to a clicking action using the Emacs-like format.
@@ -355,6 +418,8 @@ class Screen(CommandObject):
     resized to fill it. If the mode is ``"stretch"``, the image is stretched to fit all
     of it into the screen.
 
+    The ``x11_drag_polling_rate`` parameter specifies the rate for drag events in the X11 backend. By default this is set to None, indicating no limit. Because in the X11 backend we already handle motion notify events later, the performance should already be okay. However, to limit these events further you can use this variable and e.g. set it to your monitor refresh rate. 60 would mean that we handle a drag event 60 times per second.
+
     """
 
     group: _Group
@@ -368,18 +433,19 @@ class Screen(CommandObject):
         right: BarType | None = None,
         wallpaper: str | None = None,
         wallpaper_mode: str | None = None,
+        x11_drag_polling_rate: int | None = None,
         x: int | None = None,
         y: int | None = None,
         width: int | None = None,
         height: int | None = None,
     ) -> None:
-
         self.top = top
         self.bottom = bottom
         self.left = left
         self.right = right
         self.wallpaper = wallpaper
         self.wallpaper_mode = wallpaper_mode
+        self.x11_drag_polling_rate = x11_drag_polling_rate
         self.qtile: Qtile | None = None
         # x position of upper left corner can be > 0
         # if one screen is "right" of the other
@@ -407,9 +473,9 @@ class Screen(CommandObject):
         self.width = width
         self.height = height
 
-        self.set_group(group)
         for i in self.gaps:
             i._configure(qtile, self, reconfigure=reconfigure_gaps)
+        self.set_group(group)
         if self.wallpaper:
             self.wallpaper = os.path.expanduser(self.wallpaper)
             self.paint(self.wallpaper, self.wallpaper_mode)
@@ -498,7 +564,7 @@ class Screen(CommandObject):
         hook.fire("focus_change")
         hook.fire("layout_change", self.group.layouts[self.group.current_layout], self.group)
 
-    def toggle_group(self, group: _Group | None = None, warp: bool = True) -> None:
+    def _toggle_group(self, group: _Group | None = None, warp: bool = True) -> None:
         """Switch to the selected group or to the previously active one"""
         if group in (self.group, None) and self.previous_group:
             group = self.previous_group
@@ -552,6 +618,7 @@ class Screen(CommandObject):
                 return self.group if sel == self.group.name else None
         return None
 
+    @expose_command
     def resize(
         self,
         x: int | None = None,
@@ -572,29 +639,22 @@ class Screen(CommandObject):
         for bar in [self.top, self.bottom, self.left, self.right]:
             if bar:
                 bar.draw()
-        self.qtile.call_soon(self.group.layout_all)
+        self.group.layout_all()
 
-    def cmd_info(self) -> dict[str, int]:
+    @expose_command()
+    def info(self) -> dict[str, int]:
         """Returns a dictionary of info for this screen."""
         return dict(index=self.index, width=self.width, height=self.height, x=self.x, y=self.y)
 
-    def cmd_resize(
-        self,
-        x: int | None = None,
-        y: int | None = None,
-        w: int | None = None,
-        h: int | None = None,
-    ) -> None:
-        """Resize the screen"""
-        self.resize(x, y, w, h)
-
-    def cmd_next_group(self, skip_empty: bool = False, skip_managed: bool = False) -> None:
+    @expose_command()
+    def next_group(self, skip_empty: bool = False, skip_managed: bool = False) -> None:
         """Switch to the next group"""
         n = self.group.get_next_group(skip_empty, skip_managed)
         self.set_group(n)
         return n.name
 
-    def cmd_prev_group(
+    @expose_command()
+    def prev_group(
         self, skip_empty: bool = False, skip_managed: bool = False, warp: bool = True
     ) -> None:
         """Switch to the previous group"""
@@ -602,13 +662,15 @@ class Screen(CommandObject):
         self.set_group(n, warp=warp)
         return n.name
 
-    def cmd_toggle_group(self, group_name: str | None = None, warp: bool = True) -> None:
+    @expose_command()
+    def toggle_group(self, group_name: str | None = None, warp: bool = True) -> None:
         """Switch to the selected group or to the previously active one"""
         assert self.qtile is not None
         group = self.qtile.groups_map.get(group_name if group_name else "")
-        self.toggle_group(group, warp=warp)
+        self._toggle_group(group, warp=warp)
 
-    def cmd_set_wallpaper(self, path: str, mode: str | None = None) -> None:
+    @expose_command()
+    def set_wallpaper(self, path: str, mode: str | None = None) -> None:
         """Set the wallpaper to the given file."""
         self.paint(path, mode)
 
@@ -630,8 +692,8 @@ class Group:
     exclusive:
         When other apps are started in this group, should we allow them here or not?
     spawn:
-        This will be ``exec()`` d when the group is created. Tou can pass either a
-        program name or a list of programs to ``exec()``
+        This will be executed (via ``qtile.spawn()``) when the group is created. You can pass either a
+        program name or a list of programs to ``exec()``.
     layout:
         The name of default layout for this group (e.g. ``"max"``). This is the name
         specified for a particular layout in ``config.py`` or if not defined it defaults
@@ -662,7 +724,7 @@ class Group:
         exclusive: bool = False,
         spawn: str | list[str] | None = None,
         layout: str | None = None,
-        layouts: list[str] | None = None,
+        layouts: list[Layout] | None = None,
         persist: bool = True,
         init: bool = True,
         layout_opts: dict[str, Any] | None = None,
@@ -738,7 +800,6 @@ class ScratchPad(Group):
             self,
             name,
             layout="floating",
-            layouts=["floating"],
             init=False,
             position=position,
             label=label,
@@ -755,13 +816,17 @@ class ScratchPad(Group):
 
 class Match:
     """
-    Match for dynamic groups or auto-floating windows.
+    Window properties to compare (match) with a window.
 
-    It can match by title, wm_class, role, wm_type, wm_instance_class or net_wm_pid.
+    The properties will be compared to a :class:`~libqtile.base.Window` to determine if
+    its properties *match*. It can match by title, wm_class, role, wm_type,
+    wm_instance_class, net_wm_pid, or wid. Additionally, a function may be
+    passed, which takes in the :class:`~libqtile.base.Window` to be compared
+    against and returns a boolean.
 
-    :class:`Match` supports both regular expression objects (i.e. the result of
-    ``re.compile()``) or strings (match as an "include"-match). If a window matches all
-    specified values, it is considered a match.
+    For some properties, :class:`Match` supports both regular expression objects (i.e.
+    the result of ``re.compile()``) or strings (match as an "include"-match). If a
+    window matches all specified values, it is considered a match.
 
     Parameters
     ==========
@@ -781,17 +846,17 @@ class Match:
         Delegate the match to the given function, which receives the tested client as an
         argument and must return ``True`` if it matches, ``False`` otherwise.
     wid:
-        Match against the window ID.
+        Match against the window ID. This is a unique ID given to each window.
 
     """
 
     def __init__(
         self,
-        title: str | None = None,
-        wm_class: str | None = None,
-        role: str | None = None,
-        wm_type: str | None = None,
-        wm_instance_class: str | None = None,
+        title: str | re.Pattern | None = None,
+        wm_class: str | re.Pattern | None = None,
+        role: str | re.Pattern | None = None,
+        wm_type: str | re.Pattern | None = None,
+        wm_instance_class: str | re.Pattern | None = None,
         net_wm_pid: int | None = None,
         func: Callable[[base.Window], bool] | None = None,
         wid: int | None = None,
@@ -957,7 +1022,7 @@ class DropDown(configurable.Configurable):
         ),
         ("width", 0.8, "Width of window as fraction of current screen width"),
         ("height", 0.35, "Height of window as fraction of current screen."),
-        ("opacity", 0.9, "Opacity of window as fraction. Zero is opaque."),
+        ("opacity", 0.9, "Opacity of window as fraction. One is opaque."),
         (
             "on_focus_lost_hide",
             True,
@@ -981,7 +1046,7 @@ class DropDown(configurable.Configurable):
         ),
     )
 
-    def __init__(self, name: str, cmd: str, **config: dict[str, Any]) -> None:
+    def __init__(self, name: str, cmd: str, **config: Any) -> None:
         """
         Initialize :class:`DropDown` window wrapper.
 
