@@ -33,13 +33,15 @@ from shutil import which
 from typing import TYPE_CHECKING
 
 if TYPE_CHECKING:
-    from typing import Union
+    from typing import Any, Callable, Coroutine, TypeVar, Union
 
     ColorType = Union[str, tuple[int, int, int], tuple[int, int, int, float]]
     ColorsType = Union[ColorType, list[ColorType]]
 
+    T = TypeVar("T")
+
 try:
-    from dbus_next import Message, Variant
+    from dbus_next import AuthError, Message, Variant
     from dbus_next.aio import MessageBus
     from dbus_next.constants import BusType, MessageType
 
@@ -49,32 +51,50 @@ except ImportError:
 
 from libqtile.log_utils import logger
 
+# Create a list to collect references to tasks so they're not garbage collected
+# before they've run
+TASKS: list[asyncio.Task[None]] = []
+
+
+def create_task(coro: Coroutine) -> asyncio.Task | None:
+    """
+    Wrapper for asyncio.create_task.
+
+    Stores task so garbage collector doesn't remove it and removes reference when it's done.
+    See: https://textual.textualize.io/blog/2023/02/11/the-heisenbug-lurking-in-your-async-code/
+    for more info about the issue this solves.
+    """
+    loop = asyncio.get_running_loop()
+    if not loop:
+        return None
+
+    def tidy(task: asyncio.Task) -> None:
+        TASKS.remove(task)
+
+    task = asyncio.create_task(coro)
+    TASKS.append(task)
+    task.add_done_callback(tidy)
+
+    return task
+
+
+def cancel_tasks() -> None:
+    """Cancel scheduled tasks."""
+    for task in TASKS:
+        task.cancel()
+
 
 class QtileError(Exception):
     pass
 
 
-def lget(o, v):
+def lget(o: list[T], v: int) -> T | None:
     try:
         return o[v]
     except (IndexError, TypeError):
         return None
 
 
-def shuffle_up(lst):
-    if len(lst) > 1:
-        c = lst[-1]
-        lst.remove(c)
-        lst.insert(0, c)
-
-
-def shuffle_down(lst):
-    if len(lst) > 1:
-        c = lst[0]
-        lst.remove(c)
-        lst.append(c)
-
-
 def rgb(x: ColorType) -> tuple[float, float, float, float]:
     """
     Returns a valid RGBA tuple.
@@ -103,9 +123,13 @@ def rgb(x: ColorType) -> tuple[float, float, float, float]:
             alpha = float("0." + alpha_str)
         else:
             alpha = 1.0
-        if len(x) not in (6, 8):
-            raise ValueError("RGB specifier must be 6 or 8 characters long.")
-        vals = tuple(int(i, 16) for i in (x[0:2], x[2:4], x[4:6]))
+        if len(x) not in (3, 6, 8):
+            raise ValueError("RGB specifier must be 3, 6 or 8 characters long.")
+        if len(x) == 3:
+            # Multiplying by 17: 0xA * 17 = 0xAA etc.
+            vals = tuple(int(i, 16) * 17 for i in x)
+        else:
+            vals = tuple(int(i, 16) for i in (x[0:2], x[2:4], x[4:6]))
         if len(x) == 8:
             alpha = int(x[6:8], 16) / 255.0
         vals += (alpha,)  # type: ignore
@@ -113,50 +137,46 @@ def rgb(x: ColorType) -> tuple[float, float, float, float]:
     raise ValueError("Invalid RGB specifier.")
 
 
-def hex(x):
+def hex(x: ColorType) -> str:
     r, g, b, _ = rgb(x)
     return "#%02x%02x%02x" % (int(r * 255), int(g * 255), int(b * 255))
 
 
-def has_transparency(colour: ColorsType):
+def has_transparency(colour: ColorsType) -> bool:
     """
     Returns True if the colour is not fully opaque.
 
     Where a list of colours is passed, returns True if any
     colour is not fully opaque.
     """
-
-    def has_alpha(col):
-        return rgb(col)[3] < 1
-
     if isinstance(colour, (str, tuple)):
-        return has_alpha(colour)
-
-    elif isinstance(colour, list):
-        return any([has_transparency(c) for c in colour])
-
-    return False
+        return rgb(colour)[3] < 1
+    return any(has_transparency(c) for c in colour)
 
 
-def remove_transparency(colour: ColorsType):
+def remove_transparency(colour: ColorsType):  # type: ignore
     """
     Returns a tuple of (r, g, b) with no alpha.
     """
-
-    def remove_alpha(col):
-        stripped = tuple(x * 255.0 for x in rgb(col)[:3])
-        return stripped
-
     if isinstance(colour, (str, tuple)):
-        return remove_alpha(colour)
-
-    elif isinstance(colour, list):
-        return [remove_transparency(c) for c in colour]
-
-    return (0, 0, 0)
+        return tuple(x * 255.0 for x in rgb(colour)[:3])
+    return [remove_transparency(c) for c in colour]
 
 
-def scrub_to_utf8(text):
+def is_valid_colors(color: ColorsType) -> bool:
+    """
+    Returns whether the argument is a valid color or list of colors.
+    """
+    if not isinstance(color, list):
+        color = [color]
+    try:
+        list(rgb(c) for c in color)
+        return True
+    except (ValueError, TypeError):
+        return False
+
+
+def scrub_to_utf8(text: str | bytes) -> str:
     if not text:
         return ""
     elif isinstance(text, str):
@@ -165,7 +185,7 @@ def scrub_to_utf8(text):
         return text.decode("utf-8", "ignore")
 
 
-def get_cache_dir():
+def get_cache_dir() -> str:
     """
     Returns the cache directory and create if it doesn't exists
     """
@@ -180,7 +200,7 @@ def get_cache_dir():
     return cache_directory
 
 
-def describe_attributes(obj, attrs, func=lambda x: x):
+def describe_attributes(obj: Any, attrs: list[str], func: Callable = lambda x: x) -> str:
     """
     Helper for __repr__ functions to list attributes with truthy values only
     (or values that return a truthy value by func)
@@ -196,7 +216,11 @@ def describe_attributes(obj, attrs, func=lambda x: x):
     return ", ".join(pairs)
 
 
-def import_class(module_path, class_name, fallback=None):
+def import_class(
+    module_path: str,
+    class_name: str,
+    fallback: Callable | None = None,
+) -> Any:
     """Import a class safely
 
     Try to import the class module, and if it fails because of an ImporError
@@ -205,15 +229,19 @@ def import_class(module_path, class_name, fallback=None):
     try:
         module = importlib.import_module(module_path, __package__)
         return getattr(module, class_name)
-    except ImportError as error:
-        logger.warning("Unmet dependencies for '%s.%s': %s", module_path, class_name, error)
+    except ImportError:
+        logger.exception("Unmet dependencies for '%s.%s':", module_path, class_name)
         if fallback:
             logger.debug("%s", traceback.format_exc())
             return fallback(module_path, class_name)
         raise
 
 
-def lazify_imports(registry, package, fallback=None):
+def lazify_imports(
+    registry: dict[str, str],
+    package: str,
+    fallback: Callable | None = None,
+) -> tuple[tuple[str, ...], Callable, Callable]:
     """Leverage PEP 562 to make imports lazy in an __init__.py
 
     The registry must be a dictionary with the items to import as keys and the
@@ -221,10 +249,10 @@ def lazify_imports(registry, package, fallback=None):
     """
     __all__ = tuple(registry.keys())
 
-    def __dir__():
+    def __dir__() -> tuple[str, ...]:
         return __all__
 
-    def __getattr__(name):
+    def __getattr__(name: str) -> Any:
         if name not in registry:
             raise AttributeError
         module_path = "{}.{}".format(package, registry[name])
@@ -233,11 +261,17 @@ def lazify_imports(registry, package, fallback=None):
     return __all__, __dir__, __getattr__
 
 
-def send_notification(title, message, urgent=False, timeout=10000, id=None):
+def send_notification(
+    title: str,
+    message: str,
+    urgent: bool = False,
+    timeout: int = -1,
+    id_: int | None = None,
+) -> int:
     """
     Send a notification.
 
-    The id argument, if passed, requests the notification server to replace a visible
+    The id_ argument, if passed, requests the notification server to replace a visible
     notification with the same ID. An ID is returned for each call; this would then be
     passed when calling this function again to replace that notification. See:
     https://developer.gnome.org/notification-spec/
@@ -246,7 +280,7 @@ def send_notification(title, message, urgent=False, timeout=10000, id=None):
         logger.warning("dbus-next is not installed. Unable to send notifications.")
         return -1
 
-    id = randint(10, 1000) if id is None else id
+    id_ = randint(10, 1000) if id_ is None else id_
     urgency = 2 if urgent else 1
 
     try:
@@ -254,15 +288,21 @@ def send_notification(title, message, urgent=False, timeout=10000, id=None):
     except RuntimeError:
         logger.warning("Eventloop has not started. Cannot send notification.")
     else:
-        loop.create_task(_notify(title, message, urgency, timeout, id))
+        loop.create_task(_notify(title, message, urgency, timeout, id_))
 
-    return id
+    return id_
 
 
-async def _notify(title, message, urgency, timeout, id):
+async def _notify(
+    title: str,
+    message: str,
+    urgency: int,
+    timeout: int,
+    id_: int,
+) -> None:
     notification = [
         "qtile",  # Application name
-        id,  # id
+        id_,  # id
         "",  # icon
         title,  # summary
         message,  # body
@@ -282,21 +322,25 @@ async def _notify(title, message, urgency, timeout, id):
         notification,
     )
 
-    if msg.message_type == MessageType.ERROR:
-        logger.warning("Unable to send notification. " "Is a notification server running?")
+    if msg and msg.message_type == MessageType.ERROR:
+        logger.warning("Unable to send notification. Is a notification server running?")
 
     # a new bus connection is made each time a notification is sent so
     # we disconnect when the notification is done
-    bus.disconnect()
+    if bus:
+        bus.disconnect()
 
 
-def guess_terminal(preference=None):
+def guess_terminal(preference: str | Sequence | None = None) -> str | None:
     """Try to guess terminal."""
     test_terminals = []
     if isinstance(preference, str):
         test_terminals += [preference]
     elif isinstance(preference, Sequence):
         test_terminals += list(preference)
+    if "WAYLAND_DISPLAY" in os.environ:
+        # Wayland-only terminals
+        test_terminals += ["foot"]
     test_terminals += [
         "roxterm",
         "sakura",
@@ -316,22 +360,24 @@ def guess_terminal(preference=None):
         "eterm",
         "st",
         "urxvt",
+        "wezterm",
         "xterm",
         "x-terminal-emulator",
     ]
 
     for terminal in test_terminals:
-        logger.debug("Guessing terminal: {}".format(terminal))
+        logger.debug("Guessing terminal: %s", terminal)
         if not which(terminal, os.X_OK):
             continue
 
-        logger.info("Terminal found: {}".format(terminal))
+        logger.info("Terminal found: %s", terminal)
         return terminal
 
     logger.error("Default terminal has not been found.")
+    return None
 
 
-def scan_files(dirpath, *names):
+def scan_files(dirpath: str, *names: str) -> defaultdict[str, list[str]]:
     """
     Search a folder recursively for files matching those passed as arguments, with
     globbing. Returns a dict with keys equal to entries in names, and values a list of
@@ -353,8 +399,16 @@ def scan_files(dirpath, *names):
 
 
 async def _send_dbus_message(
-    session_bus, message_type, destination, interface, path, member, signature, body
-):
+    session_bus: bool,
+    message_type: MessageType,
+    destination: str | None,
+    interface: str | None,
+    path: str | None,
+    member: str | None,
+    signature: str,
+    body: Any,
+    negotiate_unix_fd: bool = False,
+) -> tuple[MessageBus | None, Message | None]:
     """
     Private method to send messages to dbus via dbus_next.
 
@@ -368,15 +422,22 @@ async def _send_dbus_message(
     if isinstance(body, str):
         body = [body]
 
-    bus = await MessageBus(bus_type=bus_type).connect()
+    try:
+        bus = await MessageBus(bus_type=bus_type, negotiate_unix_fd=negotiate_unix_fd).connect()
+    except (AuthError, Exception):
+        logger.warning("Unable to connect to dbus.")
+        return None, None
 
+    # Ignore types here: dbus-next has default values of `None` for certain
+    # parameters but the signature is `str` so passing `None` results in an
+    # error in mypy.
     msg = await bus.call(
         Message(
             message_type=message_type,
-            destination=destination,
-            interface=interface,
-            path=path,
-            member=member,
+            destination=destination,  # type: ignore
+            interface=interface,  # type: ignore
+            path=path,  # type: ignore
+            member=member,  # type: ignore
             signature=signature,
             body=body,
         )
@@ -386,18 +447,36 @@ async def _send_dbus_message(
 
 
 async def add_signal_receiver(
-    callback, session_bus=False, signal_name=None, dbus_interface=None, bus_name=None, path=None
-):
+    callback: Callable,
+    session_bus: bool = False,
+    signal_name: str | None = None,
+    dbus_interface: str | None = None,
+    bus_name: str | None = None,
+    path: str | None = None,
+    check_service: bool = False,
+) -> bool:
     """
     Helper function which aims to recreate python-dbus's add_signal_receiver
     method in dbus_next with asyncio calls.
 
+    If check_service is `True` the method will raise a wanrning and return False
+    if the service is not visible on the bus. If the `bus_name` is None, no
+    check will be performed.
+
     Returns True if subscription is successful.
     """
     if not has_dbus:
-        logger.warning("dbus-next is not installed. " "Unable to subscribe to signals")
+        logger.warning("dbus-next is not installed. Unable to subscribe to signals")
         return False
 
+    if bus_name and check_service:
+        found = await find_dbus_service(bus_name, session_bus)
+        if not found:
+            logger.warning(
+                "The %s name was not found on the bus. No callback will be attached.", bus_name
+            )
+            return False
+
     match_args = {
         "type": "signal",
         "sender": bus_name,
@@ -420,9 +499,36 @@ async def add_signal_receiver(
     )
 
     # Check if message sent successfully
-    if msg.message_type == MessageType.METHOD_RETURN:
+    if bus and msg and msg.message_type == MessageType.METHOD_RETURN:
         bus.add_message_handler(callback)
         return True
 
     else:
         return False
+
+
+async def find_dbus_service(service: str, session_bus: bool) -> bool:
+    """Looks up service name to see if it is currently available on dbus."""
+
+    # We're using low level interface here to reduce unnecessary calls for
+    # introspection etc.
+    bus, msg = await _send_dbus_message(
+        session_bus,
+        MessageType.METHOD_CALL,
+        "org.freedesktop.DBus",
+        "org.freedesktop.DBus",
+        "/org/freedesktop/DBus",
+        "ListNames",
+        "",
+        [],
+    )
+
+    if bus is None or msg is None or (msg and msg.message_type != MessageType.METHOD_RETURN):
+        logger.warning("Unable to send lookup call to dbus.")
+        return False
+
+    bus.disconnect()
+
+    names = msg.body[0]
+
+    return service in names
